/*
**  bmp2pi --- conversion from (Windows or OS/2 style) BMP to Pi
**
**  Copyright (C) 1999-2004 MIYASAKA Masaru
**
**  For conditions of distribution and use,
**  see copyright notice in common.h.
*/

#include "common.h"
#include "bmphed.h"

#define BMP2PI_VERSION		"1.00 (Sep 7, 2004)"
#define BMP2PI_COPYRIGHT	"Copyright (C) 1999-2004 MIYASAKA Masaru"

char outnam[FILENAME_MAX];
char outdir[FILENAME_MAX];
int  deletesrc = 0;
int  copytime  = 0;

#if defined(WIN32) || defined(MSDOS)
const char errlogfile[] = ".\\B2PERROR.LOG";
#else
const char errlogfile[] = "./b2perror.log";
#endif

	/* error messages */
#ifdef JAPANESE /* ---------- */
const char wrn_mkdirfail[]   =
        "WARNING: o͐fBNg܂ - %s\n"
        "WARNING:   -> -%c IvV ł̏o͐w͖܂\n";
const char err_ropenfail[]   = "SKIPPED: Yt@C܂ - %s\n";
const char err_wopenfail[]   = "SKIPPED: o̓t@C쐬ł܂ - %s\n";
const char err_outofmemory[] = "SKIPPED: Ɨpmۂł܂ - %s\n";
	/* -- */
const char err_readeof[]     = "SKIPPED: t@CrŐ؂Ă܂ - %s\n";
const char err_readerr[]     = "SKIPPED: ǂݍ݃G[܂ - %s\n";
const char err_not_a_bmp[]   = "SKIPPED: BMP t@Cł͂܂ - %s\n";
const char err_invalid_hed[] =
        "SKIPPED: BMP t@C̃wb_TCYł - %s\n";
const char err_width_zero[]  = "SKIPPED: 摜̕O(܂͕)ł - %s\n";
const char err_height_zero[] = "SKIPPED: 摜̍O(܂͕)ł - %s\n";
const char err_compression[] = "SKIPPED: sȈk^Cvł - %s\n";
const char err_invalid_bpp[] = "SKIPPED: 摜̐Fł - %s\n";
const char err_no_palette[]  = "SKIPPED: pbgĂ܂ - %s\n";
	/* -- */
const char err_writeerr[]    = "SKIPPED: ݃G[܂ - %s\n";
const char err_truecolor[]   =
      "SKIPPED: tJ[摜(Pi `ɂ)ϊł܂ - %s\n";
const char err_width_64K[]   =
      "SKIPPED:  65535 𒴂摜(Pi `ɂ)ϊł܂ - %s\n";
const char err_height_64K[]  =
      "SKIPPED:  65535 𒴂摜(Pi `ɂ)ϊł܂ - %s\n";
#else	/* ------------------- */
const char wrn_mkdirfail[]   =
        "WARNING: Cannot create a directory - %s\n"
        "WARNING:   -> Output directory specified by '-%c' will be ignored.\n";
const char err_ropenfail[]   = "SKIPPED: No such file or directory - %s\n";
const char err_wopenfail[]   = "SKIPPED: Cannot create - %s\n";
const char err_outofmemory[] = "SKIPPED: Out of memory - %s\n";
	/* -- */
const char err_readeof[]     = "SKIPPED: Premature end of BMP file - %s\n";
const char err_readerr[]     = "SKIPPED: Read operation failed - %s\n";
const char err_not_a_bmp[]   = "SKIPPED: Not a BMP file - %s\n";
const char err_invalid_hed[] = "SKIPPED: Invalid header size in BMP file - %s\n";
const char err_width_zero[]  = "SKIPPED: Invalid image width - %s\n";
const char err_height_zero[] = "SKIPPED: Invalid image height - %s\n";
const char err_compression[] = "SKIPPED: Unknown compression type - %s\n";
const char err_invalid_bpp[] = "SKIPPED: Invalid bit depth in BMP file - %s\n";
const char err_no_palette[]  = "SKIPPED: Palette is missing - %s\n";
	/* -- */
const char err_writeerr[]    = "SKIPPED: Write operation failed - %s\n";
const char err_truecolor[]   =
        "SKIPPED: Truecolor images are not supported in Pi format - %s\n";
const char err_width_64K[]   =
        "SKIPPED: Images whose width is beyond 64K are not supported - %s\n";
const char err_height_64K[]  =
        "SKIPPED: Images whose height is beyond 64K are not supported - %s\n";
#endif	/* ------------------- */

static BOOL read_bmp(char *, IMAGE *);
static const char *read_rgb_bits(IMAGE *, FILE *);
static const char *read_bitfield_bits(IMAGE *, FILE *, DWORD *, UINT);
static const char *decompress_rle_bits(IMAGE *, FILE *);
static unsigned long mgetdwl(void *);
static unsigned int mgetwl(void *);
static BOOL write_pi(char *, IMAGE *);
static const char *to8bpp(IMAGE *);
static const char *to4bpp(IMAGE *);
static void usage_exit(char *, int);



/*
**		C
*/
int main(int argc, char *argv[])
{
	char outf[FILENAME_MAX];
	IMAGE image;
	int opt;
	char *arg;
	char *p, c;
	int r_stdin, w_stdout;
	int failure = 0, success = 0;

#ifdef __LCC__		/* lcc-win32 */
	char **envp;
	void _GetMainArgs(int *, char ***, char ***, int);
	_GetMainArgs(&argc, &argv, &envp, 1);
#endif
#ifdef __EMX__
	_wildcard(&argc, &argv);
#endif
	envargv(&argc, &argv, "B2PI");

	r_stdin  = !isatty(fileno(stdin));
	w_stdout = !isatty(fileno(stdout));

	while (parsearg(&opt, &arg, argc, argv, "DdOo")) {
		switch (toupper(opt)) {
		case 'E':	deletesrc ^= 0x01;	break;
		case 'T':	copytime  ^= 0x01;	break;
		case 'Q':	quietmode ^= 0x01;	break;
		case 'L':	errorlog  ^= 0x01;	break;

		case 'X':
			r_stdin  = 0;
			w_stdout = 0;
			break;

		case 'D':				/* output directory */
			if (*arg == '-') arg = NULL;
			if (arg == NULL) {
				outdir[0] = '\0';
			} else {
				strcpy(outdir, arg);
				addslash(outdir);
				if (makedir(outdir) != 0) {
					xxprintf(wrn_mkdirfail, outdir, 'D');
					outdir[0] = '\0';
				}
			}
			break;

		case 'O':				/* output filename */
			if (arg == NULL) {
				outnam[0] = '\0';
			} else {
				strcpy(outnam, arg);
				p = basname(outnam);
				c = *p;  *p = '\0';
				if (makedir(outnam) != 0) {
					xxprintf(wrn_mkdirfail, outnam, 'O');
					outnam[0] = '\0';
				} else {
					*p = c;
				}
			}
			break;

		case 0x00:				/* input file spec */
			if (outnam[0] != '\0') {
				strcpy(outf, outnam);
				outnam[0] = '\0';
			} else if (w_stdout) {
				if (!read_bmp(arg, &image)) return 1;
				if (!write_pi(NULL, &image)) return 1;
				if (deletesrc) remove(arg);
				return 0;
			} else {
				if (outdir[0] != '\0') {
					strcat(strcpy(outf, outdir), basname(arg));
				} else {
					strcpy(outf, arg);
				}
#ifdef WIN32_LFN
				strcpy(suffix(outf), is_dos_filename(outf) ? ".PI" : ".pi");
#else
				strcpy(suffix(outf), ".pi");
#endif
			}
			/* ---------------------- */
			if (!read_bmp(arg, &image)) {
				failure++;
				break;
			}
			renbak(outf);
			if (!write_pi(outf, &image)) {
				failure++;
				break;
			}
			/* ---------------------- */
			if (copytime) cpyftime(arg, outf);
			if (deletesrc) remove(arg);
			/* ---------------------- */
			success++;
			break;

		default:
			;		/* Ignore unknown option */
		}
	}
	if (failure == 0 && success == 0) {
		if (!r_stdin) usage_exit(argv[0], 255);
		if (!read_bmp(NULL, &image)) return 1;
		if (outnam[0] != '\0') {
			renbak(outnam);
			return !write_pi(outnam, &image);
		} else if (w_stdout) {
			return !write_pi(NULL, &image);
		} else {
			strcat(strcpy(outf, outdir), "___stdin.pi");
			renbak(outf);
			return !write_pi(outf, &image);
		}
	}

	return (failure > 255) ? 255 : failure;
}


/* -----------------------------------------------------------------------
**		BMP t@C̓ǂݍ
*/

#define ERROR_ABORT(s) do { errmsg = (s); goto error_abort; } while (0)

/*
**		.bmp t@C̓ǂݍ
*/
static BOOL read_bmp(char *fn, IMAGE *img)
{
	BYTE bfh[FILEHED_SIZE + BMPV5HED_SIZE];
	BYTE *const bih = bfh + FILEHED_SIZE;
	BYTE rgbq[RGBQUAD_SIZE];
	DWORD offbits, bihsize, skip;
	DWORD compression, color_mask[3];
	UINT palette_size, true_pixdepth;
	PALETTE *pal;
	const char *errmsg;
	FILE *fp;
	int i;

	imgbuf_init(img);

	if (fn == NULL) {		/* read from stdin */
		fn = " (stdin)";
		fp = binary_stdio(fileno(stdin));
	} else {
		fp = fopen(fn, "rb");
	}
	if (fp == NULL) ERROR_ABORT(err_ropenfail);

	set_status("Reading %.80s", basname(fn));

	/* ------------------------------------------------------ */

	for (i = 0; ; i++) {		/* skip macbinary header */
		if (fread(bfh, (FILEHED_SIZE + BIHSIZE_SIZE), 1, fp) != 1)
			ERROR_ABORT(ferror(fp) ? err_readerr : err_readeof);
		if (mgetwl(bfh + BFH_WTYPE) == BMP_SIGNATURE) break;
		if (i != 0) ERROR_ABORT(err_not_a_bmp);
		if (fread(bfh, (128 - FILEHED_SIZE - BIHSIZE_SIZE), 1, fp) != 1)
			ERROR_ABORT(ferror(fp) ? err_readerr : err_readeof);
	}
	offbits = mgetdwl(bfh + BFH_DOFFBITS);
	bihsize = mgetdwl(bfh + BFH_DBIHSIZE);
	skip    = offbits - bihsize - FILEHED_SIZE;
	if (bihsize < COREHED_SIZE || bihsize > BMPV5HED_SIZE ||
	    offbits < (bihsize + FILEHED_SIZE)) ERROR_ABORT(err_invalid_hed);

	if (fread((bih + BIHSIZE_SIZE), (bihsize - BIHSIZE_SIZE), 1, fp) != 1)
		ERROR_ABORT(ferror(fp) ? err_readerr : err_readeof);

	if (bihsize >= INFOHED_SIZE) {		/* Windows-style BMP */
		img->width    = mgetdwl(bih + BIH_LWIDTH);
		img->height   = mgetdwl(bih + BIH_LHEIGHT);
		img->pixdepth = mgetwl(bih + BIH_WBITCOUNT);
		img->topdown  = FALSE;
		compression   = mgetdwl(bih + BIH_DCOMPRESSION);
		palette_size  = RGBQUAD_SIZE;
		true_pixdepth = img->pixdepth;
		if (img->height < 0) {
			img->height  = -img->height;
			img->topdown = TRUE;		/* top-down BMP */
		}
	} else {							/* OS/2-style BMP */
		img->width    = mgetwl(bih + BCH_WWIDTH);
		img->height   = mgetwl(bih + BCH_WHEIGHT);
		img->pixdepth = mgetwl(bih + BCH_WBITCOUNT);
		img->topdown  = FALSE;
		compression   = BI_RGB;
		palette_size  = RGBTRIPLE_SIZE;
		true_pixdepth = img->pixdepth;
	}
	if (img->width  <= 0) ERROR_ABORT(err_width_zero);
	if (img->height <= 0) ERROR_ABORT(err_height_zero);
	/* -- Restrictions on Pi format -- */
	if (img->width  > 0xFFFF) ERROR_ABORT(err_width_64K);
	if (img->height > 0xFFFF) ERROR_ABORT(err_height_64K);
	/* ------------------------------- */

	switch (compression) {
	case BI_RGB:
		if (img->pixdepth !=  1 && img->pixdepth !=  4 &&
		    img->pixdepth !=  8 && img->pixdepth != 16 &&
		    img->pixdepth != 24 && img->pixdepth != 32)
			ERROR_ABORT(err_invalid_bpp);

		if (img->pixdepth == 16) {
			color_mask[2] = 0x7C00;			/* red */
			color_mask[1] = 0x03E0;			/* green */
			color_mask[0] = 0x001F;			/* blue */
			compression   = BI_BITFIELDS;
		}
		break;

	case BI_BITFIELDS:
		if (img->pixdepth != 16 && img->pixdepth != 32)
			ERROR_ABORT(err_invalid_bpp);

		if (bihsize < INFOHED_SIZE + 12) {
			if (skip < (INFOHED_SIZE + 12 - bihsize))
				ERROR_ABORT(err_invalid_hed);
			if (fread((bih + bihsize), (INFOHED_SIZE + 12 - bihsize), 1, fp)
			    != 1) ERROR_ABORT(ferror(fp) ? err_readerr : err_readeof);
			skip -= (INFOHED_SIZE + 12 - bihsize);
		}
		color_mask[2] = mgetdwl(bih + B4H_DREDMASK);	/* red */
		color_mask[1] = mgetdwl(bih + B4H_DGREENMASK);	/* green */
		color_mask[0] = mgetdwl(bih + B4H_DBLUEMASK);	/* blue */

		if (img->pixdepth == 32 && color_mask[0] == 0x0000FF &&
		    color_mask[1] == 0x00FF00 && color_mask[2] == 0xFF0000) {
			compression = BI_RGB;
		}
		break;

	case BI_RLE8:
		if (img->pixdepth != 8)
			ERROR_ABORT(err_invalid_bpp);
		break;

	case BI_RLE4:
		if (img->pixdepth != 4)
			ERROR_ABORT(err_invalid_bpp);
		break;

	default:
		ERROR_ABORT(err_compression);
	}

	if (img->pixdepth == 16) img->pixdepth = 24;

	if (img->pixdepth <= 8) {
		if (skip >= palette_size << img->pixdepth) {
			img->palnum =            1 << img->pixdepth;
			skip       -= palette_size << img->pixdepth;
		} else {
			img->palnum = skip / palette_size;
			skip        = skip % palette_size;
		}
		if (img->palnum == 0)
			ERROR_ABORT(err_no_palette);
	} else {
		img->palnum = 0;
	}
	if (!imgbuf_alloc(img)) ERROR_ABORT(err_outofmemory);

	/* ------------------------------------------------------ */

	for (pal = img->palette, i = img->palnum; i > 0; pal++, i--) {
		if (fread(rgbq, palette_size, 1, fp) != 1)
			ERROR_ABORT(ferror(fp) ? err_readerr : err_readeof);
		pal->red   = rgbq[RGBQ_RED];
		pal->green = rgbq[RGBQ_GREEN];
		pal->blue  = rgbq[RGBQ_BLUE];
	}
	for ( ; skip > 0; skip--) {
		if (fgetc(fp) == EOF)
			ERROR_ABORT(ferror(fp) ? err_readerr : err_readeof);
	}

	/* ------------------------------------------------------ */

	switch (compression) {
	case BI_RGB:
		errmsg = read_rgb_bits(img, fp);
		break;
	case BI_BITFIELDS:
		errmsg = read_bitfield_bits(img, fp, color_mask, true_pixdepth);
		break;
	case BI_RLE8:
	case BI_RLE4:
		errmsg = decompress_rle_bits(img, fp);
		break;
	default:
		errmsg = err_compression;
	}
	if (errmsg != NULL) ERROR_ABORT(errmsg);

	/* ------------------------------------------------------ */

	set_status("Read OK %.80s", basname(fn));

	if (fp != stdin) fclose(fp);

	return TRUE;

error_abort:				/* error */
	xxprintf(errmsg, fn);
	if (fp != stdin && fp != NULL) fclose(fp);
	imgbuf_free(img);

	return FALSE;
}


/*
**		BI_RGB (k) `̉摜f[^ǂ
*/
static const char *read_rgb_bits(IMAGE *img, FILE *fp)
{
#if 1
	DWORD rd  = 16*1024*1024;
	DWORD num = img->imgbytes;
	BYTE *ptr = img->bmpbits;

	while (num > 0) {
		if (rd > num) rd = num;

		if (fread(ptr, rd, 1, fp) != 1)
			return ferror(fp) ? err_readerr : err_readeof;

		ptr += rd; num -= rd;
	}
#else
	if (fread(img->bmpbits, img->imgbytes, 1, fp) != 1)
		return ferror(fp) ? err_readerr : err_readeof;
#endif
	return NULL;
}


/*
**		BI_BITFIELDS `̉摜f[^ǂ
*/
static const char *read_bitfield_bits(IMAGE *img, FILE *fp, DWORD *color_mask,
                                      UINT true_pixdepth)
{
	enum { NC = 3 };
	int  color_shift[NC];
	int  color_sigbits[NC];
	BYTE color_tbl[NC][1<<7];
	DWORD true_rowbytes;
	BYTE *row, *p, *q;
	LONG w, h;
	DWORD v, u;
	int i, j, k;

	for (i = 0; i < NC; i++) {
		v = color_mask[i];
		if (v == 0) {
			color_shift[i]   = 0;
			color_sigbits[i] = 0;
			color_tbl[i][0]  = 0;
			continue;
		}
		for (j = 0; (v & 1) == 0; v >>= 1, j++) ;
		for (k = 0;     (v) != 0; v >>= 1, k++) ;
		color_shift[i]   = j;
		color_sigbits[i] = k;

		if (color_sigbits[i] <= 7) {
			k = (1 << color_sigbits[i]) - 1;
			for (j = 0; j <= k; j++)
				color_tbl[i][j] = (0xFF * j + k/2) / k;
		}
	}

	true_rowbytes = ((DWORD)img->width * (true_pixdepth/8) + 3) & (~3UL);

	for (h = img->height, row = img->bmpbits; --h >= 0;
	     row += img->rowbytes) {
		if (fread(row, true_rowbytes, 1, fp) != 1)
			return ferror(fp) ? err_readerr : err_readeof;

		switch (true_pixdepth) {
		case 16:
			for (w = img->width, p = row + (w-1)*2, q = row + (w-1)*3;
			     --w >= 0; p -= 2, q -= 3) {
				v = ((UINT)p[0]) + ((UINT)p[1] << 8);
				for (i = 0; i < NC; i++) {
					u = (v & color_mask[i]) >> color_shift[i];
					if (color_sigbits[i] <= 7)
						u = color_tbl[i][u];
					else if (color_sigbits[i] >= 9)
						u >>= (color_sigbits[i] - 8);
					q[i] = (BYTE) u;
				}
			}
			break;

		case 32:
			for (w = img->width, p = row; --w >= 0; p += 4) {
				v = ((DWORD)p[0]      ) + ((DWORD)p[1] <<  8) +
				    ((DWORD)p[2] << 16) + ((DWORD)p[3] << 24);
				for (i = 0; i < NC; i++) {
					u = (v & color_mask[i]) >> color_shift[i];
					if (color_sigbits[i] <= 7)
						u = color_tbl[i][u];
					else if (color_sigbits[i] >= 9)
						u >>= (color_sigbits[i] - 8);
					p[i] = (BYTE) u;
				}
			}
			break;
		}
	}

	return NULL;
}


/*
**		BI_RLE8/BI_RLE4 `̉摜f[^ǂ
*/
static const char *decompress_rle_bits(IMAGE *img, FILE *fp)
{
	BYTE buf[1024];		/* 258 or above */
	BYTE *bfptr = buf;
	UINT  bfcnt = 0;
	UINT  rd, reclen;
	BYTE *row = img->bmpbits;
	LONG x = 0, y = 0;
	BYTE *p, c;
	int n;

	memset(img->bmpbits, 0, img->imgbytes);

	for (;;) {
		while (bfcnt < (reclen = 2) ||
		       (bfptr[0] == 0 &&
		        ((bfptr[1] == 2 && bfcnt < (reclen += 2)) ||
		         (bfptr[1] >= 3 &&
		          bfcnt < (reclen += (bfptr[1] * img->pixdepth + 15) / 16 * 2)
		         )))) {
			if (bfptr != buf && bfcnt != 0) memmove(buf, bfptr, bfcnt);
			if ((rd = fread(buf+bfcnt, 1, sizeof(buf)-bfcnt, fp)) == 0) {
				if (x >= img->width) { /*x = 0;*/ y += 1; }
				if (y >= img->height) return NULL;	/* missing EoB marker */
				else return ferror(fp) ? err_readerr : err_readeof;
			}
			bfptr  = buf;
			bfcnt += rd;
		}
		if (y >= img->height) {
			/* We simply discard the remaining records */
			if (bfptr[0] == 0 && bfptr[1] == 1) break;	/* EoB marker */
			bfptr += reclen;
			bfcnt -= reclen;
			continue;
		}
		if (bfptr[0] != 0) {				/* Encoded-mode record */
			n = bfptr[0];  c = bfptr[1];
			switch (img->pixdepth) {
			case 8:						/* BI_RLE8 */
				while (n > 0 && x < img->width) {
					row[x] = c;
					n--; x++;
				}
				break;
			case 4:						/* BI_RLE4 */
				if (x % 2 != 0 && x < img->width) {
					c = (c >> 4) | (c << 4);
					row[x/2] = (row[x/2] & 0xF0) | (c & 0x0F);
					n--; x++;
				}
				while (n > 0 && x < img->width) {
					row[x/2] = c;
					n-=2; x+=2;
				}
				if (n < 0) x--;
				break;
			}
		} else if (bfptr[1] >= 3) {			/* Absolute-mode record */
			n = bfptr[1];  p = bfptr + 2;
			switch (img->pixdepth) {
			case 8:						/* BI_RLE8 */
				while (n > 0 && x < img->width) {
					row[x] = *p;
					n--; x++; p++;
				}
				break;
			case 4:						/* BI_RLE4 */
				if (x % 2 != 0) {
					if (x < img->width) {
						row[x/2] = (row[x/2] & 0xF0) | (*p >> 4);
						n--; x++;
					}
					while (n > 0 && x < img->width) {
						row[x/2] = (p[0] << 4) | (p[1] >> 4);
						n-=2; x+=2; p++;
					}
					if (n < 0) x--;
				} else {
					while (n > 0 && x < img->width) {
						row[x/2] = *p;
						n-=2; x+=2; p++;
					}
					if (n < 0) x--;
				}
				break;
			}
		} else if (bfptr[1] == 2) {			/* Delta record */
			x += bfptr[2]; y += bfptr[3];
			row += bfptr[3] * img->rowbytes;
		} else if (bfptr[1] == 0) {			/* End of line marker */
			x = 0; y += 1;
			row += img->rowbytes;
		} else /*if (bfptr[1] == 1)*/ {		/* End of bitmap marker */
			break;
		}
		bfptr += reclen;
		bfcnt -= reclen;
	}

	return NULL;
}


/*
**	 little-endien ` 4oCgǂ
*/
static unsigned long mgetdwl(void *ptr)
{
	unsigned char *p = ptr;

	return ((unsigned long)p[0]      ) + ((unsigned long)p[1] <<  8) +
	       ((unsigned long)p[2] << 16) + ((unsigned long)p[3] << 24);
}


/*
**	 little-endien ` 2oCgǂ
*/
static unsigned int mgetwl(void *ptr)
{
	unsigned char *p = ptr;

	return ((unsigned int)p[0]) + ((unsigned int)p[1] << 8);
}


/* -----------------------------------------------------------------------
**		Pi t@C̏
*/

/*
**		.pi t@C̏
*/
static BOOL write_pi(char *fn, IMAGE *img)
{
	static const char * const errmsgs[] = {
		NULL,				/* 0 = G[ */
		NULL,				/* 1 = .PIt@Cł͂Ȃ */
		err_truecolor,		/* 2 = ȐF[x(4,8 ȊO) */
		err_width_64K,		/* 3 = ȕ(0 or 65535) */
		err_height_64K,		/* 4 = ȍ(0 or 65535) */
		err_writeerr,		/* 5 = t@C݃G[ */
		NULL,				/* 6 = t@Cǂݍ݃G[ */
		NULL,				/* 7 = t@CI[o */
		err_outofmemory,	/* 8 = s */
	};
	pi_struct pi;
	const char *errmsg;
	FILE *fp;
	int err;

	if (fn == NULL) {
		fn = " (stdout)";
		fp = binary_stdio(fileno(stdout));
	} else {
		fp = fopen(fn, "wb");
	}
	if (fp == NULL) ERROR_ABORT(err_wopenfail);

	if ((errmsg = to8bpp(img)) != NULL) ERROR_ABORT(errmsg);
	if ((errmsg = to4bpp(img)) != NULL) ERROR_ABORT(errmsg);

	set_status("Writing %.80s", basname(fn));

	/* ------------------------------------------------------ */

	pi_write_init(&pi);

	if ((err = setjmp(pi.jmpbuf))) {
		/* If we get here, we had a problem reading the file */
		pi_write_end_io(&pi);
		pi_write_end(&pi);
		ERROR_ABORT(errmsgs[err]);
	}
	pi_write_init_io(&pi, fp);

	pi.width    = (pi_uint)img->width;
	pi.height   = (pi_uint)img->height;
	pi.bitdepth = (pi_byte)img->pixdepth;
	if (img->palnum == 0) pi.mode = PI_MODE_NO_PALETTE;
	memcpy(pi.machine, "B2PI", 4);	/* GR[_ */
/*	pi.comment.text = "Created by bmp2pi " BMP2PI_VERSION "\n";*/

	pi_write_header(&pi);

	pi_write_palette(&pi, img->palette);

	init_progress_meter(&pi);
	pi_write_image(&pi, img->rowptr);

	pi_write_end_io(&pi);
	pi_write_end(&pi);

	/* ------------------------------------------------------ */

	set_status("OK      %.80s", basname(fn));
	feed_line();

	fflush(fp);
	if (fp != stdout) fclose(fp);
	imgbuf_free(img);

	return TRUE;

error_abort:				/* error */
	if (errmsg != NULL) xxprintf(errmsg, fn);
	if (fp != stdout && fp != NULL) fclose(fp);
	imgbuf_free(img);

	return FALSE;
}


/*
**		tJ[ -> 256F ւ̌F (ordered dither)
*/
static const char *to8bpp(IMAGE *img)
{
	enum { ODITHER_SIZE = 16 };		/* dimension of dither matrix */
	static const BYTE od_matrix[ODITHER_SIZE][ODITHER_SIZE] = {
		{   0,192, 48,240, 12,204, 60,252,  3,195, 51,243, 15,207, 63,255 },
		{ 128, 64,176,112,140, 76,188,124,131, 67,179,115,143, 79,191,127 },
		{  32,224, 16,208, 44,236, 28,220, 35,227, 19,211, 47,239, 31,223 },
		{ 160, 96,144, 80,172,108,156, 92,163, 99,147, 83,175,111,159, 95 },
		{   8,200, 56,248,  4,196, 52,244, 11,203, 59,251,  7,199, 55,247 },
		{ 136, 72,184,120,132, 68,180,116,139, 75,187,123,135, 71,183,119 },
		{  40,232, 24,216, 36,228, 20,212, 43,235, 27,219, 39,231, 23,215 },
		{ 168,104,152, 88,164,100,148, 84,171,107,155, 91,167,103,151, 87 },
		{   2,194, 50,242, 14,206, 62,254,  1,193, 49,241, 13,205, 61,253 },
		{ 130, 66,178,114,142, 78,190,126,129, 65,177,113,141, 77,189,125 },
		{  34,226, 18,210, 46,238, 30,222, 33,225, 17,209, 45,237, 29,221 },
		{ 162, 98,146, 82,174,110,158, 94,161, 97,145, 81,173,109,157, 93 },
		{  10,202, 58,250,  6,198, 54,246,  9,201, 57,249,  5,197, 53,245 },
		{ 138, 74,186,122,134, 70,182,118,137, 73,185,121,133, 69,181,117 },
		{  42,234, 26,218, 38,230, 22,214, 41,233, 25,217, 37,229, 21,213 },
		{ 170,106,154, 90,166,102,150, 86,169,105,153, 89,165,101,149, 85 }
	};
		/* ---------- */
	const BYTE *od_row;
	BYTE *src, *dst;
	BYTE r, g, b, d;
	LONG x, y;
	int pixsize;

	if (img->pixdepth != 24 && img->pixdepth != 32) return NULL;

	set_status("Converting to 8bpp...");

	/* ------------------------------------------------------ */

	pixsize = img->pixdepth / 8;
	img->pixdepth = 8;

	for (y = 0; y < img->height; y++) {
		od_row = od_matrix[y % ODITHER_SIZE];
		src = dst = img->rowptr[y];
		for (x = 0; x < img->width; x++) {
			d = od_row[x % ODITHER_SIZE];
			r = (src[2] * 7 + d) / 255;
			g = (src[1] * 7 + d) / 255;
			b = (src[0] * 3 + d) / 255;
			if (r > 7) r = 7;
			if (g > 7) g = 7;
			if (b > 3) b = 3;
			dst[0] = (g << 5) + (r << 2) + b;
			src += pixsize;  dst += 1;
		}
	}

	/* ------------------------------------------------------ */

	return NULL;
}


/*
**		2F` -> 16F` ̕ϊ
*/
static const char *to4bpp(IMAGE *img)
{
	static const BYTE pix[] = { 0x00, 0x01, 0x10, 0x11 };
	IMAGE newimg;
	BYTE *src, *dst, c;
	LONG x, y;

	if (img->pixdepth != 1) return NULL;

	set_status("Converting to 4bpp...");

	/* ------------------------------------------------------ */

	newimg = *img;
	newimg.pixdepth = 4;
	newimg.palnum   = 16;

	if (!imgbuf_alloc(&newimg)) return err_outofmemory;

	memset(newimg.palette, 0, 16 * sizeof(PALETTE));
	newimg.palette[0] = img->palette[0];
	newimg.palette[1] = img->palette[1];

	for (y = 0; y < img->height; y++) {
		src =   img->rowptr[y];
		dst = newimg.rowptr[y];
		x = (img->width + 1) / 2;
		for (;;) {
			c = *src++;
			if (--x < 0) break;
			*dst++ = pix[(c >> 6) & 0x03];
			if (--x < 0) break;
			*dst++ = pix[(c >> 4) & 0x03];
			if (--x < 0) break;
			*dst++ = pix[(c >> 2) & 0x03];
			if (--x < 0) break;
			*dst++ = pix[(c >> 0) & 0x03];
		}
	}

	imgbuf_free(img);
	*img = newimg;

	/* ------------------------------------------------------ */

	return NULL;
}


/* -----------------------------------------------------------------------
**		wvXN[̕\
*/

/*
**		gp@\
*/
static void usage_exit(char *argv0, int status)
{
	static const char str_usage[] =
#ifdef JAPANESE /* -------------------------- */
#ifdef SJIS_ESCAPE
#define SJ_ESC(esc,raw)	esc
#else
#define SJ_ESC(esc,raw)	raw
#endif
		"bmp2pi, BMP -> Pi Ro[^ - version " BMP2PI_VERSION "\n"
		"   " BMP2PI_COPYRIGHT "\n"
		"\n"
		"g : %s [-XCb`] ̓t@C ...\n"
		"       : ... | %s [-XCb`] | ...\n"
		"\n"
		"̓t@Cɂ̓ChJ[hg܂ (*  ?)\n"
		"o̓t@C͓̓t@C̊gq .pi ɕςOɂȂ܂\n"
		"\n"
		"XCb`IvV (ł) :\n"
		"   -O name  o̓t@Cw肷\n"
		"   -D dir   t@Co͂fBNgw肷\n"
		"   -E       ϊꍇɂ͓̓t@C폜\n"
		"   -T       ̓t@C̃^CX^vo̓t@Cɐݒ肷\n"
		"   -Q       , ؂" SJ_ESC("\\","\") "Ȃ\n"
		"   -L       ̃G[Ot@C(%s)ɋL^\n";
#else  /* ----------------------------------- */
		"bmp2pi, a BMP-to-Pi converter - version " BMP2PI_VERSION "\n"
		"   " BMP2PI_COPYRIGHT "\n"
		"\n"
		"Usage: %s [-switches] inputfile(s) ...\n"
		"   or: ... | %s [-switches] | ...\n"
		"\n"
		"List of input files may use wildcards (* and ?)\n"
		"Output filename is same as input filename, but extension .pi\n"
		"\n"
		"Switches (case-insensitive) :\n"
		"   -O name  Specify name for output file\n"
		"   -D dir   Output files into dir\n"
		"   -E       Delete input files after successful conversion\n"
		"   -T       Set the timestamp of input file on output file\n"
		"   -Q       Quiet mode\n"
		"   -L       Log errors to %s file\n";
#endif	/* ----------------------------------- */
#if defined(WIN32) || defined(MSDOS)
	char exename[FILENAME_MAX];
	char *p;

	argv0 = strcpy(exename, basname(argv0));
	for (p = argv0; *p != '\0'; p++) *p = tolower(*p);
#endif
	fprintf(stdout, str_usage, argv0, argv0, errlogfile);

	exit(status);
}

